@@ -8,9 +8,11 @@ PDL::PP - Generate PDL routines from concise descriptions
 
 	addhdr('#include "hdr.h"');
 
+	addpm('sub foo {}');
+
 	defpdl(
 		'Transpose',
-		[qw/a(x,y,X) [o]b(y,x,X)/],
+		'a(x,y,X); int [o]b(y,x,X)',
 		'int c',
 		'loop(x,y) %{
 			$b() = $a();
@@ -22,7 +24,8 @@ PDL::PP - Generate PDL routines from concise descriptions
 =head1 DESCRIPTION
 
 This module defines the routine C<defpdl> that generates xsub code from 
-a short description such as the transpose function above.
+a short description such as the transpose function above. C<done> 
+automatically writes the files C<Prefix.xs> and C<Prefix.pm>.
 
 The idea is that since this concise description encodes in itself
 (better than C code, which would be difficult to interpret) what
@@ -49,10 +52,12 @@ this would be far superior to the usual gcc bounds checking).
 
 =head2 PDL variables
 
-The second argument to C<defpdl> is an array of strings of the form
+The second argument to C<defpdl> is either a ref to an
+array of strings of the form
 
-	[options]name(indices,X)
+	typeoption [options]name(indices,X)
 
+or a concatenation of strings like this with semicolons between them.
 Options is a comma-separated list which can at the moment contain
 
 =over 4
@@ -63,6 +68,11 @@ This pdl is used only for output and is therefore liable to be necessary
 to create at runtime. In this case, all of its indices need to have
 a defined value.
 
+=item int
+
+This pdl is of type integer and is not to be coerced to the same type
+as everything else.
+
 =back
 
 The name is a lowercase alphanumeric name for the variable. One of 
@@ -159,13 +169,7 @@ times. The outer loops should update pointers to the data
 accessed inside to be efficient. However, the comfort of writing code
 like this is very nice.
 
-At the moment,
-
-	loop(x1,x2) %{ a() = b(x => x1) * c(x => x2) %};
-
-doesn't work. 
-
-The '...' syntax does not yet work for "rest" indices.
+The current type coercion is not good.
 
 =head1 AUTHOR
 
@@ -180,7 +184,7 @@ use FileHandle;
 require Exporter;
 @ISA = qw(Exporter);
 
-@EXPORT = qw/addhdr defpdl done/;
+@EXPORT = qw/addhdr addpm defpdl done/;
 
 use Carp;
 use Data::Dumper;
@@ -191,10 +195,27 @@ sub import {
 	my ($mod,$modname, $packname, $prefix) = @_;
 	$::PDLMOD=$modname; $::PDLPACK=$packname; $::PDLPREF=$prefix;
 	$::PDLXS="";
+	$::PDLPMROUT="";
+	$::PDLPM="";
 	@_=("PDL::PP");
 	goto &Exporter::import;
 }
 
+sub addhdr {
+	my ($hdr) = @_;
+	$::PDLXS .= $hdr;
+}
+
+sub addpm {
+	my ($pm) = @_;
+	$::PDLPM .= $pm;
+}
+
+sub add_exported {
+	my ($exp) = @_;
+	$::PDLPMROUT .= $exp." ";
+}
+
 sub printxs {
 	shift;
 	$::PDLXS .= join'',@_;
@@ -205,6 +226,9 @@ sub done {
 	(my $fh = new FileHandle(">$::PDLPREF.xs")) or die "Couldn't open xs file\n";
 
 $fh->print(qq%
+/*
+ * THIS FILE WAS GENERATED BY PDL::PP! Do not modify!
+ */
 #include "EXTERN.h"
 #include "perl.h"
 #include "XSUB.h"
@@ -213,13 +237,7 @@ $fh->print(qq%
 static Core* PDL; /* Structure hold core C functions */
 SV* CoreSV;       /* Get's pointer to perl var holding core structure */
  
-%);
-
-	$fh->print($::PDLXS);
-
-$fh->print( qq%
-
-###########################################################################
+$::PDLXS
 
 BOOT:
    /* Get pointer to structure of core shared C routines */
@@ -229,9 +247,38 @@ BOOT:
    PDL = (Core*) (void*) SvIV( CoreSV );  /* Core* value */
 %);                                                                
 
+	($fh = new FileHandle(">$::PDLPREF.pm")) or die "Couldn't open pm file\n";
+
+$fh->print(qq%
+# 
+# GENERATED WITH PDL::PP! Don't modify!
+#
+package $::PDLPACK;
+
+\@EXPORT = qw( $::PDLPMROUT);
+
+use PDL::Core;
+use DynaLoader;
+\@ISA    = qw( PDL::Exporter DynaLoader ); 
+
+bootstrap $::PDLMOD;
+
+$::PDLPM;
+
+;# Exit with OK status
+
+1;
+
+%);
 
 }
 
+
+###########################################################################
+#
+#	Here is the actual code that generates the XSUB
+#
+
 sub defpdl {
 	my($name,$pdls, $others, $code) = @_;
 	my($this) = {
@@ -239,6 +286,7 @@ sub defpdl {
 		MODULE => $::PDLMOD,PACKAGE => $::PDLPACK,
 		PDLS => $pdls, PARS => $others, CODE => $code,
 	};
+	add_exported($name);
 	bless $this,PDL::PP;
 	$this->parse_pdls();
 	$this->parse_pars();
@@ -246,41 +294,63 @@ sub defpdl {
 	$this->make_parlist();
 	$this->parse_code();
 	$this->print_xsheader();
+	$this->print_xscoerce();
+	$this->printxs("\t{\n");
 	$this->print_xsdiminit();
-	$this->print_xsloopstart();
-	$this->print_xscode();
-	$this->print_xsloopend();
+	$this->print_xsgenericstart();
+	for(@{$this->get_generictypes()}) {
+		$this->print_xsgenericitem($_);
+		$this->print_xsloopstart();
+		$this->print_xscode();
+		$this->print_xsloopend();
+	}
+	$this->print_xsgenericend();
+	$this->printxs("\t}\n");
 	$this->print_xsfooter();
 }
 
 # Parse the pdl parameters and form a data structure
 sub parse_pdls { my($this) = @_;
+# If we're given a string, separate it
+	if(!ref $this->{PDLS}) {
+		$this->{PDLS} = [map {
+			/^\s*$/ ? () : $_; 	# Eliminate empties
+		} split ';',$this->{PDLS}];
+		
+	}
+# And get the info out of each one.
 	for (@{$this->{PDLS}}) {
 		/(?:
-			\[([^]]?)\]   	# $1: The initial [option] part
+			(|\w+\b)	# $1: first option
+			\[([^]]?)\]   	# $2: The initial [option] part
 	         )?
-		 ([a-z0-9]+)          	# $2: The name
-		 \(([^)]+)\)  		# $3: The indices
+		 ([a-z0-9]+)          	# $3: The name
+		 \(([^)]+)\)  		# $4: The indices
 		/x or croak "Invalid pdl def $_\n";
-#		print "PDL: '$1', '$2', '$3'\n";
-		push @{$this->{PdlNames}}, $2;
-		$this->{PdlFlags}{$2} = [split ',',$1];
-		my @inds = split ',', $3;
-		$this->{PdlInds}{$2} = [@inds];
+		print "PDL: '$1', '$2', '$3', '$4'\n";
+		my($opt1,$opt2,$name,$inds) = ($1,$2,$3,$4);
+		push @{$this->{PdlNames}}, $name;
+		$this->{PdlFlags}{$name} = [(split ',',$opt2),($opt1?$opt1:())];
+		my @inds = map{
+			s/\s//g; 		# Remove spaces
+			s/^\.\.\.$/_DOTS_/;     # ... => _DOTS_
+			$_;
+		} split ',', $inds;
+		$this->{PdlInds}{$name} = [@inds];
 		my %indcount = ();
-		$this->{PdlIndCounts}{$2} = [
+		$this->{PdlIndCounts}{$name} = [
 			map {
 				0+($indcount{$_}++);
 			} @inds
 		];
-		$this->{PdlIndTotCounts}{$2} = [
+		$this->{PdlIndTotCounts}{$name} = [
 			map {
 				($indcount{$_});
 			} @inds
 		];
 		my $ind=0;
 		for(@inds) {
-			push @{$this->{IndPdls}{$_}},[$2,$ind];
+			push @{$this->{IndPdls}{$_}},[$name,$ind];
 		}
 	}
 	@{$this->{PdlOrder}} = (
@@ -393,32 +463,43 @@ sub do_loop { my($this,$loop,$context) = @_;
 }
 
 sub do_indterm { my($this,$pdl,$ind,$subst,$context) = @_;
+	if($this->{PdlInds}{$pdl}[$_] =~ /[A-Z]/) {
+		return $pdl."_".$this->{PdlInds}{$pdl}[$_]."_offs";
+	}
+# Get informed
 	my $indname = $this->{PdlInds}{$pdl}[$ind];
 	my $indno = $this->{PdlIndCounts}{$pdl}[$ind];
 	my $indtot = $this->{PdlIndTotCounts}{$pdl}[$ind];
-	if($indtot > 1) {
-		die "Indtot > 1 not implemented!\n";
-	} else {
-		if($subst->{$indname}) {return delete $subst->{$indname};}
+# See if substitutions
+	my $substname = ($indtot>1 ? $indname.$indno : $indname);
+	my $incname = $indname.$indno;
+	my $index;
+	if($subst->{$substname}) {$index = delete $subst->{$substname};}
+	else {
+# No => get the one from the nearest context.
 		for(@$context) {
-			if($_->[0] eq $indname) {return $_->[1];}
+			if($_->[0] eq $indname) {$index = $_->[1]; break;}
 		}
 	}
+	if(!$index) {croak "Index not found: $pdl, $ind\n";}
+	return $pdl."_".$incname."_inc"."*". $index;
 }
 
 sub do_access { my($this,$access,$context) = @_;
+# Parse the access
 	$access =~ /^\$([a-z]+)\(([^)]*)\)/ or
 		croak ("Access wrong: $access\n");
 	my $pdl = $1; 
+# Parse substitutions into hash
 	my %subst = map {/^\s*(\w+)\s*=>\s*(\w*)\s*$/ or croak "Invalid subst $_\n"; ($1,$2)} split ',',$2;
-	my $text = "((double *)(${pdl}.data))"."[";
+# Generate the text
+	my $text = "(${pdl}_datap)"."[";
 	$text .= join '+',map {
-		$this->{PdlInds}{$pdl}[$_] =~ /[A-Z]/ ?
-		 $pdl."_".$this->{PdlInds}{$pdl}[$_]."_offs" :
-	 	$pdl."_".$this->{PdlInds}{$pdl}[$_]."_inc".
-		"*".$this->do_indterm($pdl,$_,\%subst,$context);
+		$this->do_indterm($pdl,$_,\%subst,$context);
 	} (0..$#{$this->{PdlInds}{$pdl}});
 	$text .= "]";
+# If not all substitutions made, the user probably made a spelling
+# error. Barf.
 	if(scalar(keys %subst) != 0) {
 		croak("Substitutions left: ".(join ',',keys %subst)."\n");
 	}
@@ -464,15 +545,22 @@ sub print_xsdiminit { my($this) = @_;
 		if(!$this->{IndCheck}{$_}) {
 			$this->printxs("\tint ${_}_size = -1, ${_}_incs = -1 ;\n");
 		}
-		$this->printxs("\tint ".(join',',map {"${_}_${ind}_inc"}
-			    @{$this->{PdlNames}}).";\n");
 	}
-	for (@{$this->{PdlNames}}) {
-		$this->printxs("\tint ${_}_ndims;\n");
-	}
-	for (@{$this->{PdlNames}}) {
-		$this->printxs(qq#\tif(${_}.datatype != PDL_D) {die("NON-DOUBLE!\n");}\n#);
+	my $pdl;
+	for $pdl (@{$this->{PdlNames}}) {
+# Make the index increments. Here, if the same index repeats, all repetitions
+# need to have their increments.
+		my $ind;
+		$this->printxs("\tint ".(join',',
+		   map {$ind=$_; 
+		         "${pdl}_$this->{PdlInds}{$pdl}[$ind]$this->{PdlIndCounts}{$pdl}[$ind]_inc"
+		   	    }
+			    0..$#{$this->{PdlInds}{$pdl}}).";\n");
+		$this->printxs("\tint ${pdl}_ndims;\n");
 	}
+#	for (@{$this->{PdlNames}}) {
+#		$this->printxs(qq#\tif(${_}.datatype != PDL_D) {croak("NON-DOUBLE!\n");}\n#);
+#	}
 # Now, loop over the matrices, filling in the dimensions.
 # Check at compile-time to see when the rest dimensions are defined.
 	my %restdef=();
@@ -505,7 +593,7 @@ sub print_xsdiminit { my($this) = @_;
 				carp "Undef restind not last!\n";
 			}
 			$this->printxs(
-			 qq#\tif(${_}_ndims > ${_}.ndims) {die("Too many dims for $_\\n");}\n#
+			 qq#\tif(${_}_ndims > ${_}.ndims) {croak("Too many dims for $_\\n");}\n#
 			);
 			$this->printxs(
 			 "\t$undef[0]_ndims = ${_}.ndims - ${_}_ndims;
@@ -522,7 +610,7 @@ sub print_xsdiminit { my($this) = @_;
 			$restdef{$undef[0]} = 1;
 		} else {
 			$this->printxs(
-			qq#\tif(${_}_ndims != ${_}.ndims) {die("Dimensions unequal!\\n");}\n#);
+			qq#\tif(${_}_ndims != ${_}.ndims) {croak("Dimensions unequal!\\n");}\n#);
 		}
 # Test the dimension sizes.
 # Alternatively, if this pdl is not defined, create the matrix
@@ -533,18 +621,21 @@ sub print_xsdiminit { my($this) = @_;
 # Now, test whether the indices are equal.
 		$this->printxs("\t__flag=0;__mult=1;${pdl}_ndims=0;\n");
 		my $a;
+		my $no=0;
 		$this->printxs(join '',
 		  map {
+		    $no++;
 		    /[A-Z]/ ? "\tfor(__ind=0; __ind<${_}_ndims; __ind++) {
 		       ${pdl}_${_}_incs[__ind] = __mult; __mult *= ${_}_sizes[__ind];
 		       if(${pdl}.dims[${pdl}_ndims++] != ${_}_sizes[__ind]) 
 		       	{__flag++;}}\n"
 		       :"
 		        if(${_}_size == -1) {${_}_size = ${pdl}.dims[${pdl}_ndims];}
-		        ${pdl}_${_}_inc = __mult; __mult *= ${_}_size;
+		        ${pdl}_${_}$this->{PdlIndCounts}{$pdl}[$no-1]_inc 
+				= __mult; __mult *= ${_}_size;
 		        if(${pdl}.dims[${pdl}_ndims++] != ${_}_size) {__flag++;}\n"}
 		    @{$this->{PdlInds}{$_}});
-		$this->printxs(qq#\tif(__flag) {die("Some dimensions not equal!\\n");}\n#);
+		$this->printxs(qq#\tif(__flag) {croak("Some dimensions not equal!\\n");}\n#);
 
 	}
 }
@@ -576,14 +667,67 @@ sub print_xscode { my($this) = @_;
 # End the loop
 sub print_xsloopend { my($this) = @_;
 	$this->printxs("\t/* Ending loop over rest-dimensions */\n");
+	$this->{ContLoopInd}++;
+	my $contloop = "contloop".$this->{ContLoopInd};
 	for(keys %{$this->{RestInds}}) {
 		$this->printxs(
-		 qq#\tfor(__ind=0; __ind<${_}_ndims; __ind++) {if(++(${_}_inds[__ind]) >= ${_}_sizes[__ind]) {${_}_inds[__ind]=0; break;} else goto contloop;} \n#);
+		 qq#\tfor(__ind=0; __ind<${_}_ndims; __ind++) {if(++(${_}_inds[__ind]) >= ${_}_sizes[__ind]) {${_}_inds[__ind]=0; break;} else goto $contloop;} \n#);
 	}
-	$this->printxs("\t__restend=1;\n\tcontloop: 1;}\n");
+	$this->printxs("\t__restend=1;\n\t$contloop: 1;}\n");
 }
 
 sub print_xsfooter { my($this) = @_;
 	$this->printxs("\t}\n");
 }
 
+####
+#
+# Type coercion
+
+sub print_xscoerce { my($this) = @_;
+	$this->printxs("\tint __datatype=PDL_B;\n");
+	for(@{$this->{PdlNames}}) {
+		if(!grep {/^int$/} @{$this->{PdlFlags}{$_}}) {
+			$this->printxs("\tif($_.datatype > __datatype)
+				__datatype = $_.datatype;\n");
+		} else {
+			print "Not doing $_: has int type\n";
+			$this->printxs(qq%\tif($_.datatype != PDL_L) croak("Invalid datatype for $_: should be long\n");%);
+		}
+	}
+	for(@{$this->{PdlNames}}) {
+		if(!grep {/^int$/} @{$this->{PdlFlags}{$_}}) {
+			$this->printxs("\tPDL->converttype(&($_),__datatype,1);\n");
+		}
+	}
+}
+
+sub print_xsgenericstart { my($this) = @_;
+	$this->printxs("/* Start generic loop */\n");
+	$this->printxs("\tswitch(__datatype) { case -42: /* Warning eater */ {1;\n");
+}
+
+sub print_xsgenericitem { my($this,$item) = @_;
+	$this->printxs("\t} break; case $item->[0]: {\n");
+	for(@{$this->{PdlNames}}) {
+		if(!grep {/^int$/} @{$this->{PdlFlags}{$_}}) {
+			$this->printxs("\t$item->[1] *${_}_datap = ($item->[1] *)${_}.data;\n");
+		} else {
+			$this->printxs("\tint *${_}_datap = (int *)${_}.data;\n");
+		}
+	}
+}
+
+sub print_xsgenericend { my($this) = @_;
+	$this->printxs("\t}}\n");
+}
+
+sub get_generictypes {
+	return [[PDL_B,"unsigned char"],
+		[PDL_S,"short"],
+		[PDL_US,"unsigned short"],
+		[PDL_L,"long"],
+		[PDL_F,"float"],
+		[PDL_D,"double"]];
+}
+
